@@ -1,4 +1,4 @@
-# Copyright (C) 2014-2020 all contributors <meta@public-inbox.org>
+# Copyright (C) all contributors <meta@public-inbox.org>
 # License: AGPL-3.0+ <https://www.gnu.org/licenses/agpl-3.0.txt>
 #
 # Used throughout the project for reading configuration
@@ -12,27 +12,37 @@ use strict;
 use v5.10.1;
 use PublicInbox::Inbox;
 use PublicInbox::Spawn qw(popen_rd);
+our $LD_PRELOAD = $ENV{LD_PRELOAD}; # only valid at startup
+our $DEDUPE; # set to {} to dedupe or clear cache
 
 sub _array ($) { ref($_[0]) eq 'ARRAY' ? $_[0] : [ $_[0] ] }
 
 # returns key-value pairs of config directives in a hash
 # if keys may be multi-value, the value is an array ref containing all values
 sub new {
-	my ($class, $file) = @_;
+	my ($class, $file, $errfh) = @_;
 	$file //= default_file();
 	my $self;
+	my $set_dedupe;
 	if (ref($file) eq 'SCALAR') { # used by some tests
 		open my $fh, '<', $file or die;  # PerlIO::scalar
 		$self = config_fh_parse($fh, "\n", '=');
+		bless $self, $class;
 	} else {
-		$self = git_config_dump($file);
+		if (-f $file && $DEDUPE) {
+			$file = rel2abs_collapsed($file);
+			$self = $DEDUPE->{$file} and return $self;
+			$set_dedupe = 1;
+		}
+		$self = git_config_dump($class, $file, $errfh);
+		$self->{'-f'} = $file;
 	}
-	bless $self, $class;
 	# caches
 	$self->{-by_addr} = {};
 	$self->{-by_list_id} = {};
 	$self->{-by_name} = {};
 	$self->{-by_newsgroup} = {};
+	$self->{-by_eidx_key} = {};
 	$self->{-no_obfuscate} = {};
 	$self->{-limiters} = {};
 	$self->{-code_repos} = {}; # nick => PublicInbox::Git object
@@ -59,7 +69,7 @@ sub new {
 	if (my $css = delete $self->{'publicinbox.css'}) {
 		$self->{css} = _array($css);
 	}
-
+	$DEDUPE->{$file} = $self if $set_dedupe;
 	$self;
 }
 
@@ -86,9 +96,22 @@ sub lookup_list_id {
 
 sub lookup_name ($$) {
 	my ($self, $name) = @_;
-	$self->{-by_name}->{$name} // _fill($self, "publicinbox.$name");
+	$self->{-by_name}->{$name} // _fill_ibx($self, $name);
 }
 
+sub lookup_ei {
+	my ($self, $name) = @_;
+	$self->{-ei_by_name}->{$name} //= _fill_ei($self, $name);
+}
+
+sub lookup_eidx_key {
+	my ($self, $eidx_key) = @_;
+	_lookup_fill($self, '-by_eidx_key', $eidx_key);
+}
+
+# special case for [extindex "all"]
+sub ALL { lookup_ei($_[0], 'all') }
+
 sub each_inbox {
 	my ($self, $cb, @arg) = @_;
 	# may auto-vivify if config file is non-existent:
@@ -123,20 +146,16 @@ sub default_file {
 
 sub config_fh_parse ($$$) {
 	my ($fh, $rs, $fs) = @_;
-	my %rv;
-	my (%section_seen, @section_order);
+	my (%rv, %seen, @section_order, $line, $k, $v, $section, $cur, $i);
 	local $/ = $rs;
-	while (defined(my $line = <$fh>)) {
-		chomp $line;
-		my ($k, $v) = split($fs, $line, 2);
-		my ($section) = ($k =~ /\A(\S+)\.[^\.]+\z/);
-		unless (defined $section_seen{$section}) {
-			$section_seen{$section} = 1;
-			push @section_order, $section;
-		}
+	while (defined($line = <$fh>)) { # perf critical with giant configs
+		$i = index($line, $fs);
+		$k = substr($line, 0, $i);
+		$v = substr($line, $i + 1, -1); # chop off $fs
+		$section = substr($k, 0, rindex($k, '.'));
+		$seen{$section} //= push(@section_order, $section);
 
-		my $cur = $rv{$k};
-		if (defined $cur) {
+		if (defined($cur = $rv{$k})) {
 			if (ref($cur) eq "ARRAY") {
 				push @$cur, $v;
 			} else {
@@ -152,18 +171,17 @@ sub config_fh_parse ($$$) {
 }
 
 sub git_config_dump {
-	my ($file) = @_;
-	return {} unless -e $file;
-	my @cmd = (qw/git config -z -l --includes/, "--file=$file");
-	my $cmd = join(' ', @cmd);
-	my $fh = popen_rd(\@cmd);
+	my ($class, $file, $errfh) = @_;
+	return bless {}, $class unless -e $file;
+	my $cmd = [ qw(git config -z -l --includes), "--file=$file" ];
+	my $fh = popen_rd($cmd, undef, { 2 => $errfh // 2 });
 	my $rv = config_fh_parse($fh, "\0", "\n");
-	close $fh or die "failed to close ($cmd) pipe: $?";
-	$rv;
+	close $fh or die "@$cmd failed: \$?=$?\n";
+	bless $rv, $class;
 }
 
-sub valid_inbox_name ($) {
-	my ($name) = @_;
+sub valid_foo_name ($;$) {
+	my ($name, $pfx) = @_;
 
 	# Similar rules found in git.git/remote.c::valid_remote_nick
 	# and git.git/refs.c::check_refname_component
@@ -171,6 +189,7 @@ sub valid_inbox_name ($) {
 	if ($name eq '' || $name =~ /\@\{/ ||
 	    $name =~ /\.\./ || $name =~ m![/:\?\[\]\^~\s\f[:cntrl:]\*]! ||
 	    $name =~ /\A\./ || $name =~ /\.\z/) {
+		warn "invalid $pfx name: `$name'\n" if $pfx;
 		return 0;
 	}
 
@@ -311,44 +330,29 @@ sub parse_cgitrc {
 	cgit_repo_merge($self, $repo->{dir}, $repo) if $repo;
 }
 
-# parse a code repo
-# Only git is supported at the moment, but SVN and Hg are possibilities
-sub _fill_code_repo {
+# parse a code repo, only git is supported at the moment
+sub fill_code_repo {
 	my ($self, $nick) = @_;
 	my $pfx = "coderepo.$nick";
-
-	# TODO: support gitweb and other repository viewers?
-	if (defined(my $cgitrc = delete $self->{-cgitrc_unparsed})) {
-		parse_cgitrc($self, $cgitrc, 0);
-	}
-	my $dir = $self->{"$pfx.dir"}; # aka "GIT_DIR"
-	unless (defined $dir) {
+	my $dir = $self->{"$pfx.dir"} // do { # aka "GIT_DIR"
 		warn "$pfx.dir unset\n";
 		return;
-	}
-
+	};
 	my $git = PublicInbox::Git->new($dir);
-	foreach my $t (qw(blob commit tree tag)) {
-		$git->{$t.'_url_format'} =
-				_array($self->{lc("$pfx.${t}UrlFormat")});
-	}
-
 	if (defined(my $cgits = $self->{"$pfx.cgiturl"})) {
 		$git->{cgit_url} = $cgits = _array($cgits);
 		$self->{"$pfx.cgiturl"} = $cgits;
-
-		# cgit supports "/blob/?id=%s", but it's only a plain-text
-		# display and requires an unabbreviated id=
-		foreach my $t (qw(blob commit tag)) {
-			$git->{$t.'_url_format'} //= map {
-				"$_/$t/?id=%s"
-			} @$cgits;
-		}
 	}
 
 	$git;
 }
 
+sub get_all {
+	my ($self, $key) = @_;
+	my $v = $self->{$key} // return;
+	_array($v);
+}
+
 sub git_bool {
 	my ($val) = $_[-1]; # $_[0] may be $self, or $val
 	if ($val =~ /\A(?:false|no|off|[\-\+]?(?:0x)?0+)\z/i) {
@@ -360,37 +364,73 @@ sub git_bool {
 	}
 }
 
-sub _fill {
-	my ($self, $pfx) = @_;
-	my $ibx = {};
+# abs_path resolves symlinks, so we want to avoid it if rel2abs
+# is sufficient and doesn't leave "/.." or "/../"
+sub rel2abs_collapsed {
+	require File::Spec;
+	my $p = File::Spec->rel2abs($_[-1]);
+	return $p if substr($p, -3, 3) ne '/..' && index($p, '/../') < 0;
+	require Cwd;
+	Cwd::abs_path($p);
+}
 
-	for my $k (qw(watch nntpserver)) {
+sub get_1 {
+	my ($self, $key) = @_;
+	my $v = $self->{$key};
+	return $v if !ref($v);
+	warn "W: $key has multiple values, only using `$v->[-1]'\n";
+	$v->[-1];
+}
+
+sub repo_objs {
+	my ($self, $ibxish) = @_;
+	my $ibx_code_repos = $ibxish->{coderepo} // return;
+	$ibxish->{-repo_objs} // do {
+		my $code_repos = $self->{-code_repos};
+		my @repo_objs;
+		for my $nick (@$ibx_code_repos) {
+			my @parts = split(m!/!, $nick);
+			for (@parts) {
+				@parts = () unless valid_foo_name($_);
+			}
+			unless (@parts) {
+				warn "invalid coderepo name: `$nick'\n";
+				next;
+			}
+			my $repo = $code_repos->{$nick} //=
+						fill_code_repo($self, $nick);
+			push @repo_objs, $repo if $repo;
+		}
+		if (scalar @repo_objs) {
+			$ibxish ->{-repo_objs} = \@repo_objs;
+		} else {
+			delete $ibxish->{coderepo};
+		}
+	}
+}
+
+sub _fill_ibx {
+	my ($self, $name) = @_;
+	my $pfx = "publicinbox.$name";
+	my $ibx = {};
+	for my $k (qw(watch)) {
 		my $v = $self->{"$pfx.$k"};
 		$ibx->{$k} = $v if defined $v;
 	}
 	for my $k (qw(filter inboxdir newsgroup replyto httpbackendmax feedmax
-			indexlevel indexsequentialshard)) {
-		if (defined(my $v = $self->{"$pfx.$k"})) {
-			if (ref($v) eq 'ARRAY') {
-				warn <<EOF;
-W: $pfx.$k has multiple values, only using `$v->[-1]'
-EOF
-				$ibx->{$k} = $v->[-1];
-			} else {
-				$ibx->{$k} = $v;
-			}
-		}
+			indexlevel indexsequentialshard boost)) {
+		my $v = get_1($self, "$pfx.$k") // next;
+		$ibx->{$k} = $v;
 	}
 
-	# backwards compatibility:
-	$ibx->{inboxdir} //= $self->{"$pfx.mainrepo"};
-	if (($ibx->{inboxdir} // '') =~ /\n/s) {
-		warn "E: `$ibx->{inboxdir}' must not contain `\\n'\n";
+	# "mainrepo" is backwards compatibility:
+	my $dir = $ibx->{inboxdir} //= $self->{"$pfx.mainrepo"} // return;
+	if (index($dir, "\n") >= 0) {
+		warn "E: `$dir' must not contain `\\n'\n";
 		return;
 	}
-	foreach my $k (qw(obfuscate)) {
-		my $v = $self->{"$pfx.$k"};
-		defined $v or next;
+	for my $k (qw(obfuscate)) {
+		my $v = $self->{"$pfx.$k"} // next;
 		if (defined(my $bval = git_bool($v))) {
 			$ibx->{$k} = $bval;
 		} else {
@@ -399,24 +439,17 @@ EOF
 	}
 	# TODO: more arrays, we should support multi-value for
 	# more things to encourage decentralization
-	foreach my $k (qw(address altid nntpmirror coderepo hide listid url
-			infourl watchheader)) {
-		if (defined(my $v = $self->{"$pfx.$k"})) {
-			$ibx->{$k} = _array($v);
-		}
-	}
-
-	return unless defined($ibx->{inboxdir});
-	my $name = $pfx;
-	$name =~ s/\Apublicinbox\.//;
-
-	if (!valid_inbox_name($name)) {
-		warn "invalid inbox name: '$name'\n";
-		return;
+	for my $k (qw(address altid nntpmirror imapmirror
+			coderepo hide listid url
+			infourl watchheader
+			nntpserver imapserver pop3server)) {
+		my $v = $self->{"$pfx.$k"} // next;
+		$ibx->{$k} = _array($v);
 	}
 
+	return unless valid_foo_name($name, 'publicinbox');
 	$ibx->{name} = $name;
-	$ibx->{-pi_config} = $self;
+	$ibx->{-pi_cfg} = $self;
 	$ibx = PublicInbox::Inbox->new($ibx);
 	foreach (@{$ibx->{address}}) {
 		my $lc_addr = lc($_);
@@ -429,8 +462,31 @@ EOF
 			$self->{-by_list_id}->{lc($list_id)} = $ibx;
 		}
 	}
-	if (my $ng = $ibx->{newsgroup}) {
-		$self->{-by_newsgroup}->{$ng} = $ibx;
+	if (defined(my $ngname = $ibx->{newsgroup})) {
+		if (ref($ngname)) {
+			delete $ibx->{newsgroup};
+			warn 'multiple newsgroups not supported: '.
+				join(', ', @$ngname). "\n";
+		# Newsgroup name needs to be compatible with RFC 3977
+		# wildmat-exact and RFC 3501 (IMAP) ATOM-CHAR.
+		# Leave out a few chars likely to cause problems or conflicts:
+		# '|', '<', '>', ';', '#', '$', '&',
+		} elsif ($ngname =~ m![^A-Za-z0-9/_\.\-\~\@\+\=:]! ||
+				$ngname eq '') {
+			delete $ibx->{newsgroup};
+			warn "newsgroup name invalid: `$ngname'\n";
+		} else {
+			# PublicInbox::NNTPD does stricter ->nntp_usable
+			# checks, keep this lean for startup speed
+			$self->{-by_newsgroup}->{$ngname} = $ibx;
+		}
+	}
+	unless (defined $ibx->{newsgroup}) { # for ->eidx_key
+		my $abs = rel2abs_collapsed($dir);
+		if ($abs ne $dir) {
+			warn "W: `$dir' canonicalized to `$abs'\n";
+			$ibx->{inboxdir} = $abs;
+		}
 	}
 	$self->{-by_name}->{$name} = $ibx;
 	if ($ibx->{obfuscate}) {
@@ -438,30 +494,42 @@ EOF
 		$ibx->{-no_obfuscate_re} = $self->{-no_obfuscate_re};
 		fill_all($self); # noop to populate -no_obfuscate
 	}
-
-	if (my $ibx_code_repos = $ibx->{coderepo}) {
-		my $code_repos = $self->{-code_repos};
-		my $repo_objs = $ibx->{-repo_objs} = [];
-		foreach my $nick (@$ibx_code_repos) {
-			my @parts = split(m!/!, $nick);
-			my $valid = 0;
-			$valid += valid_inbox_name($_) foreach (@parts);
-			$valid == scalar(@parts) or next;
-
-			my $repo = $code_repos->{$nick} //=
-						_fill_code_repo($self, $nick);
-			push @$repo_objs, $repo if $repo;
-		}
+	if (my $es = ALL($self)) {
+		require PublicInbox::Isearch;
+		$ibx->{isrch} = PublicInbox::Isearch->new($ibx, $es);
 	}
+	$self->{-by_eidx_key}->{$ibx->eidx_key} = $ibx;
+}
 
-	$ibx
+sub _fill_ei ($$) {
+	my ($self, $name) = @_;
+	eval { require PublicInbox::ExtSearch } or return;
+	my $pfx = "extindex.$name";
+	my $d = $self->{"$pfx.topdir"} // return;
+	-d $d or return;
+	if (index($d, "\n") >= 0) {
+		warn "E: `$d' must not contain `\\n'\n";
+		return;
+	}
+	my $es = PublicInbox::ExtSearch->new($d);
+	for my $k (qw(indexlevel indexsequentialshard)) {
+		my $v = get_1($self, "$pfx.$k") // next;
+		$es->{$k} = $v;
+	}
+	for my $k (qw(coderepo hide url infourl)) {
+		my $v = $self->{"$pfx.$k"} // next;
+		$es->{$k} = _array($v);
+	}
+	return unless valid_foo_name($name, 'extindex');
+	$es->{name} = $name;
+	$es;
 }
 
 sub urlmatch {
 	my ($self, $key, $url) = @_;
 	state $urlmatch_broken; # requires git 1.8.5
 	return if $urlmatch_broken;
-	my $file = default_file();
+	my $file = $self->{'-f'} // default_file();
 	my $cmd = [qw/git config -z --includes --get-urlmatch/,
 		"--file=$file", $key, $url ];
 	my $fh = popen_rd($cmd);
@@ -476,4 +544,25 @@ sub urlmatch {
 	}
 }
 
+sub json {
+	state $json;
+	$json //= do {
+		for my $mod (qw(Cpanel::JSON::XS JSON::MaybeXS JSON JSON::PP)) {
+			eval "require $mod" or next;
+			# ->ascii encodes non-ASCII to "\uXXXX"
+			$json = $mod->new->ascii(1) and last;
+		}
+		$json;
+	};
+}
+
+sub squote_maybe ($) {
+	my ($val) = @_;
+	if ($val =~ m{([^\w@\./,\%\+\-])}) {
+		$val =~ s/(['!])/'\\$1'/g; # '!' for csh
+		return "'$val'";
+	}
+	$val;
+}
+
 1;
