@@ -1,89 +1,157 @@
-# Copyright (C) 2015-2020 all contributors <meta@public-inbox.org>
+# Copyright (C) all contributors <meta@public-inbox.org>
 # License: AGPL-3.0+ <https://www.gnu.org/licenses/agpl-3.0.txt>
-# contains common daemon code for the httpd, imapd, and nntpd servers.
-# This may be used for read-only IMAP server if we decide to implement it.
+#
+# Contains common daemon code for the httpd, imapd, and nntpd servers
+# and designed for handling thousands of untrusted clients over slow
+# and/or lossy connections.
 package PublicInbox::Daemon;
 use strict;
-use warnings;
-use Getopt::Long qw/:config gnu_getopt no_ignore_case auto_abbrev/;
+use v5.10.1;
+use Getopt::Long qw(:config gnu_getopt no_ignore_case auto_abbrev);
 use IO::Handle; # ->autoflush
 use IO::Socket;
+use File::Spec;
 use POSIX qw(WNOHANG :signal_h);
 use Socket qw(IPPROTO_TCP SOL_SOCKET);
-sub SO_ACCEPTFILTER () { 0x1000 }
-use Cwd qw/abs_path/;
 STDOUT->autoflush(1);
 STDERR->autoflush(1);
 use PublicInbox::DS qw(now);
-use PublicInbox::Syscall qw($SFD_NONBLOCK);
-require PublicInbox::Listener;
+use PublicInbox::Listener;
 use PublicInbox::EOFpipe;
 use PublicInbox::Sigfd;
+use PublicInbox::Git;
+use PublicInbox::GitAsyncCat;
+use PublicInbox::Eml;
+use PublicInbox::Config;
+our $SO_ACCEPTFILTER = 0x1000;
 my @CMD;
 my ($set_user, $oldset);
 my (@cfg_listen, $stdout, $stderr, $group, $user, $pid_file, $daemonize);
 my $worker_processes = 1;
 my @listeners;
-my %pids;
-my %tls_opt; # scheme://sockname => args for IO::Socket::SSL->start_SSL
+my (%pids, %logs);
+my %tls_opt; # scheme://sockname => args for IO::Socket::SSL::SSL_Context->new
 my $reexec_pid;
 my ($uid, $gid);
 my ($default_cert, $default_key);
-my %KNOWN_TLS = ( 443 => 'https', 563 => 'nntps', 993 => 'imaps' );
-my %KNOWN_STARTTLS = ( 119 => 'nntp', 143 => 'imap' );
+my %KNOWN_TLS = (443 => 'https', 563 => 'nntps', 993 => 'imaps', 995 =>'pop3s');
+my %KNOWN_STARTTLS = (110 => 'pop3', 119 => 'nntp', 143 => 'imap');
+my %SCHEME2PORT = map { $KNOWN_TLS{$_} => $_ + 0 } keys %KNOWN_TLS;
+for (keys %KNOWN_STARTTLS) { $SCHEME2PORT{$KNOWN_STARTTLS{$_}} = $_ + 0 }
+$SCHEME2PORT{http} = 80;
 
-sub accept_tls_opt ($) {
-	my ($opt_str) = @_;
-	# opt_str: opt1=val1,opt2=val2 (opt may repeat for multi-value)
-	require PublicInbox::TLS;
+sub listener_opt ($) {
+	my ($str) = @_; # opt1=val1,opt2=val2 (opt may repeat for multi-value)
 	my $o = {};
 	# allow ',' as delimiter since '&' is shell-unfriendly
-	foreach (split(/[,&]/, $opt_str)) {
+	for (split(/[,&]/, $str)) {
 		my ($k, $v) = split(/=/, $_, 2);
-		push @{$o->{$k} ||= []}, $v;
+		push @{$o->{$k}}, $v;
 	}
 
 	# key may be a part of cert.  At least
 	# p5-io-socket-ssl/example/ssl_server.pl has this fallback:
-	$o->{cert} //= [ $default_cert ];
+	$o->{cert} //= [ $default_cert ] if defined($default_cert);
 	$o->{key} //= defined($default_key) ? [ $default_key ] : $o->{cert};
-	my %ctx_opt = (SSL_server => 1);
+	$o;
+}
+
+sub check_absolute ($$) {
+	my ($var, $val) = @_;
+	die <<EOM if index($val // '/', '/') != 0;
+$var must be an absolute path when using --daemonize: $val
+EOM
+}
+
+sub accept_tls_opt ($) {
+	my ($opt) = @_;
+	my $o = ref($opt) eq 'HASH' ? $opt : listener_opt($opt);
+	return if !defined($o->{cert});
+	require PublicInbox::TLS;
+	my @ctx_opt;
 	# parse out hostname:/path/to/ mappings:
-	foreach my $k (qw(cert key)) {
-		my $x = $ctx_opt{'SSL_'.$k.'_file'} = {};
+	for my $k (qw(cert key)) {
+		$o->{$k} // next;
+		push(@ctx_opt, "SSL_${k}_file", {});
 		foreach my $path (@{$o->{$k}}) {
 			my $host = '';
 			$path =~ s/\A([^:]+):// and $host = $1;
-			$x->{$host} = $path;
+			$ctx_opt[-1]->{$host} = $path;
+			check_absolute($k, $path) if $daemonize;
 		}
 	}
-	my $ctx = IO::Socket::SSL::SSL_Context->new(%ctx_opt) or
-		die 'SSL_Context->new: '.PublicInbox::TLS::err();
-
-	# save ~34K per idle connection (cf. SSL_CTX_set_mode(3ssl))
-	# RSS goes from 346MB to 171MB with 10K idle NNTPS clients on amd64
-	# cf. https://rt.cpan.org/Ticket/Display.html?id=129463
-	my $mode = eval { Net::SSLeay::MODE_RELEASE_BUFFERS() };
-	if ($mode && $ctx->{context}) {
-		eval { Net::SSLeay::CTX_set_mode($ctx->{context}, $mode) };
-		warn "W: $@ (setting SSL_MODE_RELEASE_BUFFERS)\n" if $@;
-	}
-
-	{ SSL_server => 1, SSL_startHandshake => 0, SSL_reuse_ctx => $ctx };
+	\@ctx_opt;
 }
 
-sub daemon_prepare ($) {
-	my ($default_listen) = @_;
+sub do_chown ($) {
+	$uid // return;
+	my ($path) = @_;
+	chown($uid, $gid, $path) or warn "chown $path: $!\n";
+}
+
+sub open_log_path ($$) { # my ($fh, $path) = @_; # $_[0] is modified
+	open $_[0], '>>', $_[1] or die "open(>> $_[1]): $!";
+	$_[0]->autoflush(1);
+	do_chown($_[1]);
+	$_[0];
+}
+
+sub load_mod ($;$$) {
+	my ($scheme, $opt, $addr) = @_;
+	my $modc = "PublicInbox::\U$scheme";
+	$modc =~ s/S\z//;
+	my $mod = $modc.'D';
+	eval "require $mod"; # IMAPD|HTTPD|NNTPD|POP3D
+	die $@ if $@;
+	my %xn;
+	my $tlsd = $xn{tlsd} = $mod->new;
+	my %env = map {
+		substr($_, length('env.')) => $opt->{$_}->[-1];
+	} grep(/\Aenv\./, keys %$opt);
+	$xn{refresh} = sub {
+		my ($sig) = @_;
+		local @ENV{keys %env} = values %env;
+		$tlsd->refresh_groups($sig);
+	};
+	$xn{post_accept} = $tlsd->can('post_accept_cb') ?
+			$tlsd->post_accept_cb : sub { $modc->new($_[0], $tlsd) };
+	my @paths = qw(out err);
+	if ($modc eq 'PublicInbox::HTTP') {
+		@paths = qw(err);
+		$xn{af_default} = 'httpready';
+		if (my $p = $opt->{psgi}) {
+			die "multiple psgi= options specified\n" if @$p > 1;
+			check_absolute('psgi=', $p->[0]) if $daemonize;
+			$tlsd->{psgi} = $p->[0];
+			warn "# $scheme://$addr psgi=$p->[0]\n";
+		}
+	}
+	for my $f (@paths) {
+		my $p = $opt->{$f} or next;
+		die "multiple $f= options specified\n" if @$p > 1;
+		check_absolute("$f=", $p->[0]) if $daemonize;
+		$p = File::Spec->canonpath($p->[0]);
+		$tlsd->{$f} = $logs{$p} //= open_log_path(my $fh, $p);
+		warn "# $scheme://$addr $f=$p\n";
+	}
+	my $err = $tlsd->{err};
+	$tlsd->{warn_cb} = sub { print $err @_ }; # for local $SIG{__WARN__}
+	\%xn;
+}
+
+sub daemon_prepare ($$) {
+	my ($default_listen, $xnetd) = @_;
 	my $listener_names = {}; # sockname => IO::Handle
-	$oldset = PublicInbox::Sigfd::block_signals();
+	$oldset = PublicInbox::DS::block_signals();
 	@CMD = ($0, @ARGV);
 	my ($prog) = ($CMD[0] =~ m!([^/]+)\z!g);
+	my $dh = defined($default_listen) ? " (default: $default_listen)" : '';
 	my $help = <<EOF;
 usage: $prog [-l ADDRESS] [--cert=FILE] [--key=FILE]
 
 options:
 
-  -l ADDRESS    address to listen on (default: $default_listen)
+  -l ADDRESS    address to listen on$dh
   --cert=FILE   default SSL/TLS certificate
   --key=FILE    default SSL/TLS certificate
   -W WORKERS    number of worker processes to spawn (default: 1)
@@ -106,39 +174,50 @@ EOF
 	GetOptions(%opt) or die $help;
 	if ($show_help) { print $help; exit 0 };
 
+	$_ = File::Spec->canonpath($_ // next) for ($stdout, $stderr);
 	if (defined $pid_file && $pid_file =~ /\.oldbin\z/) {
 		die "--pid-file cannot end with '.oldbin'\n";
 	}
 	@listeners = inherit($listener_names);
-
-	# allow socket-activation users to set certs once and not
-	# have to configure each socket:
-	my @inherited_names = keys(%$listener_names) if defined($default_cert);
+	my @inherited_names = keys(%$listener_names);
 
 	# ignore daemonize when inheriting
 	$daemonize = undef if scalar @listeners;
 
-	push @cfg_listen, $default_listen unless (@listeners || @cfg_listen);
-
+	unless (@listeners || @cfg_listen) {
+		$default_listen // die "no listeners specified\n";
+		push @cfg_listen, $default_listen
+	}
+	my ($default_scheme) = (($default_listen // '') =~ m!\A([^:]+)://!);
 	foreach my $l (@cfg_listen) {
 		my $orig = $l;
-		my $scheme = '';
-		if ($l =~ s!\A([^:]+)://!!) {
-			$scheme = $1;
-		} elsif ($l =~ /\A(?:\[[^\]]+\]|[^:]+):([0-9])+/) {
-			my $s = $KNOWN_TLS{$1} // $KNOWN_STARTTLS{$1};
-			$scheme = $s if defined $s;
+		my ($scheme, $port, $opt);
+		$l =~ s!\A([a-z0-9]+)://!! and $scheme = $1;
+		$scheme //= $default_scheme;
+		if ($l =~ /\A(?:\[[^\]]+\]|[^:]+):([0-9]+)/) {
+			$port = $1 + 0;
+			$scheme //= $KNOWN_TLS{$port} // $KNOWN_STARTTLS{$port};
 		}
+		$scheme // die "unable to determine URL scheme of $orig\n";
+		if (!defined($port) && index($l, '/') != 0) { # AF_UNIX socket
+			$port = $SCHEME2PORT{$scheme} //
+				die "no port in listen=$orig\n";
+			$l =~ s!\A([^/]+)!$1:$port! or
+				die "unable to add port=$port to $l\n";
+		}
+		$l =~ s!/\z!!; # chop one trailing slash
 		if ($l =~ s!/?\?(.+)\z!!) {
-			$tls_opt{"$scheme://$l"} = accept_tls_opt($1);
+			$opt = listener_opt($1);
+			$tls_opt{"$scheme://$l"} = accept_tls_opt($opt);
 		} elsif (defined($default_cert)) {
 			$tls_opt{"$scheme://$l"} = accept_tls_opt('');
-		} elsif ($scheme =~ /\A(?:https|imaps|imaps)\z/) {
+		} elsif ($scheme =~ /\A(?:https|imaps|nntps|pop3s)\z/) {
 			die "$orig specified w/o cert=\n";
 		}
-		# TODO: use scheme to load either NNTP.pm or HTTP.pm
-
-		next if $listener_names->{$l}; # already inherited
+		if ($listener_names->{$l}) { # already inherited
+			$xnetd->{$l} = load_mod($scheme, $opt, $l);
+			next;
+		}
 		my (%o, $sock_pkg);
 		if (index($l, '/') == 0) {
 			$sock_pkg = 'IO::Socket::UNIX';
@@ -165,50 +244,58 @@ EOF
 		}
 		$o{Listen} = 1024;
 		my $prev = umask 0000;
-		my $s = eval { $sock_pkg->new(%o) };
-		warn "error binding $l: $! ($@)\n" unless $s;
+		my $s = eval { $sock_pkg->new(%o) } or
+			warn "error binding $l: $! ($@)\n";
 		umask $prev;
-		if ($s) {
-			$listener_names->{sockname($s)} = $s;
-			$s->blocking(0);
-			push @listeners, $s;
-		}
+		$s // next;
+		$s->blocking(0);
+		my $sockname = sockname($s);
+		warn "# bound $scheme://$sockname\n";
+		$xnetd->{$sockname} //= load_mod($scheme);
+		$listener_names->{$sockname} = $s;
+		push @listeners, $s;
 	}
 
 	# cert/key options in @cfg_listen takes precedence when inheriting,
 	# but map well-known inherited ports if --listen isn't specified
-	# at all
-	for my $sockname (@inherited_names) {
-		$sockname =~ /:([0-9]+)\z/ or next;
-		if (my $scheme = $KNOWN_TLS{$1}) {
-			$tls_opt{"$scheme://$sockname"} ||= accept_tls_opt('');
-		} elsif (($scheme = $KNOWN_STARTTLS{$1})) {
-			next if $tls_opt{"$scheme://$sockname"};
-			$tls_opt{''} ||= accept_tls_opt('');
+	# at all.  This allows socket-activation users to set certs once
+	# and not have to configure each socket:
+	if (defined $default_cert) {
+		my ($stls) = (($default_scheme // '') =~ /\A(pop3|nntp|imap)/);
+		for my $x (@inherited_names) {
+			$x =~ /:([0-9]+)\z/ or next; # no TLS for AF_UNIX
+			if (my $scheme = $KNOWN_TLS{$1}) {
+				$xnetd->{$x} //= load_mod($scheme);
+				$tls_opt{"$scheme://$x"} ||= accept_tls_opt('');
+			} elsif (($scheme = $KNOWN_STARTTLS{$1})) {
+				$xnetd->{$x} //= load_mod($scheme);
+				$tls_opt{"$scheme://$x"} ||= accept_tls_opt('');
+			} elsif (defined $stls) {
+				$tls_opt{"$stls://$x"} ||= accept_tls_opt('');
+			}
 		}
 	}
-
-	die "No listeners bound\n" unless @listeners;
-}
-
-sub check_absolute ($$) {
-	my ($var, $val) = @_;
-	if (defined $val && index($val, '/') != 0) {
-		die
-"--$var must be an absolute path when using --daemonize: $val\n";
+	if (defined $default_scheme) {
+		for my $x (@inherited_names) {
+			$xnetd->{$x} //= load_mod($default_scheme);
+		}
 	}
+	die "No listeners bound\n" unless @listeners;
 }
 
 sub daemonize () {
 	if ($daemonize) {
+		require Cwd;
 		foreach my $i (0..$#ARGV) {
 			my $arg = $ARGV[$i];
 			next unless -e $arg;
-			$ARGV[$i] = abs_path($arg);
+			$ARGV[$i] = Cwd::abs_path($arg);
 		}
-		check_absolute('stdout', $stdout);
-		check_absolute('stderr', $stderr);
-		check_absolute('pid-file', $pid_file);
+		check_absolute('--stdout', $stdout);
+		check_absolute('--stderr', $stderr);
+		check_absolute('--pid-file', $pid_file);
+		check_absolute('--cert', $default_cert);
+		check_absolute('--key', $default_key);
 
 		chdir '/' or die "chdir failed: $!";
 	}
@@ -236,8 +323,7 @@ EOF
 	};
 
 	if ($daemonize) {
-		my $pid = fork;
-		die "could not fork: $!\n" unless defined $pid;
+		my $pid = fork // die "fork: $!";
 		exit if $pid;
 
 		open(STDIN, '+<', '/dev/null') or
@@ -245,8 +331,7 @@ EOF
 		open STDOUT, '>&STDIN' or die "redirect stdout failed: $!\n";
 		open STDERR, '>&STDIN' or die "redirect stderr failed: $!\n";
 		POSIX::setsid();
-		$pid = fork;
-		die "could not fork: $!\n" unless defined $pid;
+		$pid = fork // die "fork: $!";
 		exit if $pid;
 	}
 	return unless defined $pid_file;
@@ -269,13 +354,11 @@ sub worker_quit { # $_[0] = signal name or number (unused)
 		my ($dmap, undef) = @_;
 		my $n = 0;
 		my $now = now();
-
-		foreach my $s (values %$dmap) {
+		for my $s (values %$dmap) {
 			$s->can('busy') or next;
-			if ($s->busy($now)) {
+			if ($s->busy) {
 				++$n;
-			} else {
-				# close as much as possible, early as possible
+			} else { # close as much as possible, early as possible
 				$s->close;
 			}
 		}
@@ -295,18 +378,9 @@ sub worker_quit { # $_[0] = signal name or number (unused)
 }
 
 sub reopen_logs {
-	if ($stdout) {
-		open STDOUT, '>>', $stdout or
-			warn "failed to redirect stdout to $stdout: $!\n";
-		STDOUT->autoflush(1);
-		do_chown($stdout);
-	}
-	if ($stderr) {
-		open STDERR, '>>', $stderr or
-			warn "failed to redirect stderr to $stderr: $!\n";
-		STDERR->autoflush(1);
-		do_chown($stderr);
-	}
+	$logs{$stdout} //= \*STDOUT if defined $stdout;
+	$logs{$stderr} //= \*STDERR if defined $stderr;
+	while (my ($p, $fh) = each %logs) { open_log_path($fh, $p) }
 }
 
 sub sockname ($) {
@@ -366,17 +440,16 @@ sub inherit ($) {
 	my $end = $fds + 2; # LISTEN_FDS_START - 1
 	my @rv = ();
 	foreach my $fd (3..$end) {
-		my $s = IO::Handle->new_from_fd($fd, 'r');
+		open(my $s, '<&=', $fd) or warn "fdopen fd=$fd: $!";
 		if (my $k = sockname($s)) {
-			if ($s->blocking) {
-				$s->blocking(0);
-				warn <<"";
-Inherited socket (fd=$fd) is blocking, making it non-blocking.
+			my $prev_was_blocking = $s->blocking(0);
+			warn <<"" if $prev_was_blocking;
+Inherited socket ($k fd=$fd) is blocking, making it non-blocking.
 Set 'NonBlocking = true' in the systemd.service unit to avoid stalled
 processes when multiple service instances start.
 
-			}
 			$listener_names->{$k} = $s;
+			warn "# inherited $k fd=$fd\n";
 			push @rv, $s;
 		} else {
 			warn "failed to inherit fd=$fd (LISTEN_FDS=$fds)";
@@ -422,11 +495,8 @@ sub upgrade { # $_[0] = signal name or number (unused)
 }
 
 sub kill_workers ($) {
-	my ($s) = @_;
-
-	while (my ($pid, $id) = each %pids) {
-		kill $s, $pid;
-	}
+	my ($sig) = @_;
+	kill $sig, keys(%pids);
 }
 
 sub upgrade_aborted ($) {
@@ -517,9 +587,9 @@ EOF
 		},
 		CHLD => \&reap_children,
 	};
-	my $sigfd = PublicInbox::Sigfd->new($sig, 0);
-	local %SIG = (%SIG, %$sig) if !$sigfd;
-	PublicInbox::Sigfd::sig_setmask($oldset) if !$sigfd;
+	my $sigfd = PublicInbox::Sigfd->new($sig);
+	local @SIG{keys %$sig} = values(%$sig) unless $sigfd;
+	PublicInbox::DS::sig_setmask($oldset) if !$sigfd;
 	while (1) { # main loop
 		my $n = scalar keys %pids;
 		unless (@listeners) {
@@ -535,12 +605,15 @@ EOF
 		}
 		my $want = $worker_processes - 1;
 		if ($n <= $want) {
-			PublicInbox::Sigfd::block_signals() if !$sigfd;
+			PublicInbox::DS::block_signals() if !$sigfd;
 			for my $i ($n..$want) {
+				my $seed = rand(0xffffffff);
 				my $pid = fork;
 				if (!defined $pid) {
 					warn "failed to fork worker[$i]: $!\n";
 				} elsif ($pid == 0) {
+					srand($seed);
+					eval { Net::SSLeay::randomize() };
 					$set_user->() if $set_user;
 					return $p0; # run normal work code
 				} else {
@@ -548,7 +621,7 @@ EOF
 					$pids{$pid} = $i;
 				}
 			}
-			PublicInbox::Sigfd::sig_setmask($oldset) if !$sigfd;
+			PublicInbox::DS::sig_setmask($oldset) if !$sigfd;
 		}
 
 		if ($sigfd) { # Linux and IO::KQueue users:
@@ -560,12 +633,11 @@ EOF
 	exit # never gets here, just for documentation
 }
 
-sub tls_start_cb ($$) {
-	my ($opt, $orig_post_accept) = @_;
+sub tls_cb {
+	my ($post_accept, $tlsd) = @_;
 	sub {
 		my ($io, $addr, $srv) = @_;
-		my $ssl = IO::Socket::SSL->start_SSL($io, %$opt);
-		$orig_post_accept->($ssl, $addr, $srv);
+		$post_accept->(PublicInbox::TLS::start($io, $tlsd), $addr, $srv)
 	}
 }
 
@@ -580,23 +652,35 @@ sub defer_accept ($$) {
 		return if $sec > 0; # systemd users may set a higher value
 		setsockopt($s, IPPROTO_TCP, $TCP_DEFER_ACCEPT, 1);
 	} elsif ($^O eq 'freebsd') {
-		my $x = getsockopt($s, SOL_SOCKET, SO_ACCEPTFILTER);
+		my $x = getsockopt($s, SOL_SOCKET, $SO_ACCEPTFILTER);
 		return if defined $x; # don't change if set
 		my $accf_arg = pack('a16a240', $af_name, '');
-		setsockopt($s, SOL_SOCKET, SO_ACCEPTFILTER, $accf_arg);
+		setsockopt($s, SOL_SOCKET, $SO_ACCEPTFILTER, $accf_arg);
 	}
 }
 
-sub daemon_loop ($$$$) {
-	my ($refresh, $post_accept, $tlsd, $af_default) = @_;
-	my %post_accept;
-	while (my ($k, $v) = each %tls_opt) {
-		if ($k =~ s!\A(?:https|imaps|nntps)://!!) {
-			$post_accept{$k} = tls_start_cb($v, $post_accept);
-		} elsif ($tlsd) { # STARTTLS, $k eq '' is OK
-			$tlsd->{accept_tls} = $v;
+sub daemon_loop ($) {
+	my ($xnetd) = @_;
+	local $PublicInbox::Config::DEDUPE = {}; # enable dedupe cache
+	my $refresh = sub {
+		my ($sig) = @_;
+		%$PublicInbox::Config::DEDUPE = (); # clear cache
+		for my $xn (values %$xnetd) {
+			delete $xn->{tlsd}->{ssl_ctx}; # PublicInbox::TLS::start
+			eval { $xn->{refresh}->($sig) };
+			warn "refresh $@\n" if $@;
 		}
+	};
+	my %post_accept;
+	while (my ($k, $ctx_opt) = each %tls_opt) {
+		$ctx_opt // next;
+		my ($scheme, $l) = split(m!://!, $k, 2);
+		my $xn = $xnetd->{$l} // die "BUG: no xnetd for $k";
+		$xn->{tlsd}->{ssl_ctx_opt} //= $ctx_opt;
+		$scheme =~ m!\A(?:https|imaps|nntps|pop3s)! and
+			$post_accept{$l} = tls_cb(@$xn{qw(post_accept tlsd)});
 	}
+	undef %tls_opt;
 	my $sig = {
 		HUP => $refresh,
 		INT => \&worker_quit,
@@ -622,44 +706,35 @@ sub daemon_loop ($$$$) {
 	$uid = $gid = undef;
 	reopen_logs();
 	@listeners = map {;
-		my $tls_cb = $post_accept{sockname($_)};
+		my $l = sockname($_);
+		my $tls_cb = $post_accept{$l};
+		my $xn = $xnetd->{$l} // die "BUG: no xnetd for $l";
 
 		# NNTPS, HTTPS, HTTP, IMAPS and POP3S are client-first traffic
 		# IMAP, NNTP and POP3 are server-first
-		defer_accept($_, $tls_cb ? 'dataready' : $af_default);
+		defer_accept($_, $tls_cb ? 'dataready' : $xn->{af_default});
 
 		# this calls epoll_create:
-		PublicInbox::Listener->new($_, $tls_cb || $post_accept)
+		PublicInbox::Listener->new($_, $tls_cb || $xn->{post_accept})
 	} @listeners;
-	my $sigfd = PublicInbox::Sigfd->new($sig, $SFD_NONBLOCK);
-	local %SIG = (%SIG, %$sig) if !$sigfd;
-	if (!$sigfd) {
-		# wake up every second to accept signals if we don't
-		# have signalfd or IO::KQueue:
-		PublicInbox::Sigfd::sig_setmask($oldset);
-		PublicInbox::DS->SetLoopTimeout(1000);
-	}
-	PublicInbox::DS->EventLoop;
+	PublicInbox::DS::event_loop($sig, $oldset);
 }
 
-sub run ($$$;$) {
-	my ($default, $refresh, $post_accept, $tlsd) = @_;
-	local $SIG{PIPE} = 'IGNORE';
-	daemon_prepare($default);
-	my $af_default = $default =~ /:8080\z/ ? 'httpready' : undef;
+sub run {
+	my ($default_listen) = @_;
+	daemon_prepare($default_listen, my $xnetd = {});
 	my $for_destroy = daemonize();
-	daemon_loop($refresh, $post_accept, $tlsd, $af_default);
+
+	# localize GCF2C for tests:
+	local $PublicInbox::GitAsyncCat::GCF2C;
+	local $PublicInbox::Git::async_warn = 1;
+	local $SIG{__WARN__} = PublicInbox::Eml::warn_ignore_cb();
+
+	daemon_loop($xnetd);
 	PublicInbox::DS->Reset;
 	# ->DESTROY runs when $for_destroy goes out-of-scope
 }
 
-sub do_chown ($) {
-	my ($path) = @_;
-	if (defined $uid and !chown($uid, $gid, $path)) {
-		warn "could not chown $path: $!\n";
-	}
-}
-
 sub write_pid ($) {
 	my ($path) = @_;
 	Net::Server::Daemonize::create_pid_file($path);
