@@ -1,4 +1,4 @@
-# Copyright (C) 2016-2020 all contributors <meta@public-inbox.org>
+# Copyright (C) all contributors <meta@public-inbox.org>
 # License: AGPL-3.0+ <https://www.gnu.org/licenses/agpl-3.0.txt>
 #
 # git fast-import-based ssoma-mda MDA replacement
@@ -9,7 +9,7 @@ package PublicInbox::Import;
 use strict;
 use parent qw(PublicInbox::Lock);
 use v5.10.1;
-use PublicInbox::Spawn qw(spawn popen_rd);
+use PublicInbox::Spawn qw(run_die popen_rd);
 use PublicInbox::MID qw(mids mid2path);
 use PublicInbox::Address;
 use PublicInbox::Smsg;
@@ -19,13 +19,23 @@ use PublicInbox::MDA;
 use PublicInbox::Eml;
 use POSIX qw(strftime);
 
+sub default_branch () {
+	state $default_branch = do {
+		my $r = popen_rd([qw(git config --global init.defaultBranch)],
+				 { GIT_CONFIG => undef });
+		chomp(my $h = <$r> // '');
+		close $r;
+		$h eq '' ? 'refs/heads/master' : "refs/heads/$h";
+	}
+}
+
 sub new {
 	# we can't change arg order, this is documented in POD
 	# and external projects may rely on it:
 	my ($class, $git, $name, $email, $ibx) = @_;
-	my $ref = 'refs/heads/master';
+	my $ref;
 	if ($ibx) {
-		$ref = $ibx->{ref_head} // 'refs/heads/master';
+		$ref = $ibx->{ref_head};
 		$name //= $ibx->{name};
 		$email //= $ibx->{-primary_address};
 		$git //= $ibx->git;
@@ -34,7 +44,7 @@ sub new {
 		git => $git,
 		ident => "$name <$email>",
 		mark => 1,
-		ref => $ref,
+		ref => $ref // default_branch,
 		ibx => $ibx,
 		path_type => '2/38', # or 'v2'
 		lock_path => "$git->{git_dir}/ssoma.lock", # v2 changes this
@@ -46,9 +56,9 @@ sub new {
 sub gfi_start {
 	my ($self) = @_;
 
-	return ($self->{in}, $self->{out}) if $self->{pid};
+	return ($self->{in}, $self->{out}) if $self->{in};
 
-	my (@ret, $out_r, $out_w);
+	my ($in_r, $out_r, $out_w);
 	pipe($out_r, $out_w) or die "pipe failed: $!";
 
 	$self->lock_acquire;
@@ -56,27 +66,24 @@ sub gfi_start {
 		my ($git, $ref) = @$self{qw(git ref)};
 		local $/ = "\n";
 		chomp($self->{tip} = $git->qx(qw(rev-parse --revs-only), $ref));
+		die "fatal: rev-parse --revs-only $ref: \$?=$?" if $?;
 		if ($self->{path_type} ne '2/38' && $self->{tip}) {
-			local $/ = "\0";
-			my @t = $git->qx(qw(ls-tree -r -z --name-only), $ref);
-			chomp @t;
-			$self->{-tree} = { map { $_ => 1 } @t };
+			my $t = $git->qx(qw(ls-tree -r -z --name-only), $ref);
+			die "fatal: ls-tree -r -z --name-only $ref: \$?=$?" if $?;
+			$self->{-tree} = { map { $_ => 1 } split(/\0/, $t) };
 		}
-		my @cmd = ('git', "--git-dir=$git->{git_dir}",
-			qw(fast-import --quiet --done --date-format=raw));
-		my ($in_r, $pid) = popen_rd(\@cmd, undef, { 0 => $out_r });
+		$in_r = $self->{in} = $git->popen(qw(fast-import
+					--quiet --done --date-format=raw),
+					undef, { 0 => $out_r });
 		$out_w->autoflush(1);
-		$self->{in} = $in_r;
 		$self->{out} = $out_w;
-		$self->{pid} = $pid;
 		$self->{nchg} = 0;
-		@ret = ($in_r, $out_w);
 	};
 	if ($@) {
 		$self->lock_release;
 		die $@;
 	}
-	@ret;
+	($in_r, $out_w);
 }
 
 sub wfail () { die "write to fast-import failed: $!" }
@@ -96,7 +103,7 @@ sub _check_path ($$$$) {
 	return if $tip eq '';
 	print $w "ls $tip $path\n" or wfail;
 	local $/ = "\n";
-	defined(my $info = <$r>) or die "EOF from fast-import: $!";
+	my $info = <$r> // die "EOF from fast-import: $!";
 	$info =~ /\Amissing / ? undef : $info;
 }
 
@@ -104,22 +111,21 @@ sub _cat_blob ($$$) {
 	my ($r, $w, $oid) = @_;
 	print $w "cat-blob $oid\n" or wfail;
 	local $/ = "\n";
-	my $info = <$r>;
-	defined $info or die "EOF from fast-import / cat-blob: $!";
+	my $info = <$r> // die "EOF from fast-import / cat-blob: $!";
 	$info =~ /\A[a-f0-9]{40,} blob ([0-9]+)\n\z/ or return;
 	my $left = $1;
 	my $offset = 0;
 	my $buf = '';
 	my $n;
 	while ($left > 0) {
-		$n = read($r, $buf, $left, $offset);
-		defined($n) or die "read cat-blob failed: $!";
+		$n = read($r, $buf, $left, $offset) //
+			die "read cat-blob failed: $!";
 		$n == 0 and die 'fast-export (cat-blob) died';
 		$left -= $n;
 		$offset += $n;
 	}
-	$n = read($r, my $lf, 1);
-	defined($n) or die "read final byte of cat-blob failed: $!";
+	$n = read($r, my $lf, 1) //
+		die "read final byte of cat-blob failed: $!";
 	die "bad read on final byte: <$lf>" if $lf ne "\n";
 
 	# fixup some bugginess in old versions:
@@ -141,10 +147,8 @@ sub check_remove_v1 {
 	my $oid = $1;
 	my $msg = _cat_blob($r, $w, $oid) or die "BUG: cat-blob $1 failed";
 	my $cur = PublicInbox::Eml->new($msg);
-	my $cur_s = $cur->header('Subject');
-	$cur_s = '' unless defined $cur_s;
-	my $cur_m = $mime->header('Subject');
-	$cur_m = '' unless defined $cur_m;
+	my $cur_s = $cur->header('Subject') // '';
+	my $cur_m = $mime->header('Subject') // '';
 	if ($cur_s ne $cur_m || norm_body($cur) ne norm_body($mime)) {
 		return ('MISMATCH', $cur);
 	}
@@ -153,14 +157,14 @@ sub check_remove_v1 {
 
 sub checkpoint {
 	my ($self) = @_;
-	return unless $self->{pid};
+	return unless $self->{in};
 	print { $self->{out} } "checkpoint\n" or wfail;
 	undef;
 }
 
 sub progress {
 	my ($self, $msg) = @_;
-	return unless $self->{pid};
+	return unless $self->{in};
 	print { $self->{out} } "progress $msg\n" or wfail;
 	readline($self->{in}) eq "progress $msg\n" or die
 		"progress $msg not received\n";
@@ -209,10 +213,10 @@ sub barrier {
 # used for v2
 sub get_mark {
 	my ($self, $mark) = @_;
-	die "not active\n" unless $self->{pid};
+	die "not active\n" unless $self->{in};
 	my ($r, $w) = $self->gfi_start;
 	print $w "get-mark $mark\n" or wfail;
-	defined(my $oid = <$r>) or die "get-mark failed, need git 2.6.0+\n";
+	my $oid = <$r> // die "get-mark failed, need git 2.6.0+\n";
 	chomp($oid);
 	$oid;
 }
@@ -404,8 +408,22 @@ sub add {
 	# v2: we need this for Xapian
 	if ($smsg) {
 		$smsg->{blob} = $self->get_mark(":$blob");
-		$smsg->{raw_bytes} = $n;
-		$smsg->{-raw_email} = \$raw_email;
+		$smsg->set_bytes($raw_email, $n);
+		if (my $oidx = delete $smsg->{-oidx}) { # used by LeiStore
+			my $eidx_git = delete $smsg->{-eidx_git};
+
+			# we need this sharedkv to dedupe blobs added in the
+			# same fast-import transaction
+			my $u = $self->{uniq_skv} //= do {
+				require PublicInbox::SharedKV;
+				my $x = PublicInbox::SharedKV->new;
+				$x->dbh;
+				$x;
+			};
+			return if !$u->set_maybe($smsg->oidbin, 1);
+			return if (!$oidx->vivify_xvmd($smsg) &&
+					$eidx_git->check($smsg->{blob}));
+		}
 	}
 	my $ref = $self->{ref};
 	my $commit = $self->{mark}++;
@@ -429,17 +447,7 @@ sub add {
 	$self->{tip} = ":$commit";
 }
 
-sub run_die ($;$$) {
-	my ($cmd, $env, $rdr) = @_;
-	my $pid = spawn($cmd, $env, $rdr);
-	waitpid($pid, 0) == $pid or die join(' ', @$cmd) .' did not finish';
-	$? == 0 or die join(' ', @$cmd) . " failed: $?\n";
-}
-
-my @INIT_FILES = ('HEAD' => "ref: refs/heads/master\n",
-		'description' => <<EOD,
-Unnamed repository; edit this file 'description' to name the repository.
-EOD
+my @INIT_FILES = ('HEAD' => undef, # filled in at runtime
 		'config' => <<EOC);
 [core]
 	repositoryFormatVersion = 0
@@ -450,15 +458,18 @@ EOD
 EOC
 
 sub init_bare {
-	my ($dir) = @_; # or self
+	my ($dir, $head) = @_; # or self
 	$dir = $dir->{git}->{git_dir} if ref($dir);
 	require File::Path;
 	File::Path::mkpath([ map { "$dir/$_" } qw(objects/info refs/heads) ]);
-	for (my $i = 0; $i < @INIT_FILES; $i++) {
-		my $f = $dir.'/'.$INIT_FILES[$i++];
+	$INIT_FILES[1] //= 'ref: '.default_branch."\n";
+	my @fn_contents = @INIT_FILES;
+	$fn_contents[1] = "ref: refs/heads/$head\n" if defined $head;
+	while (my ($fn, $contents) = splice(@fn_contents, 0, 2)) {
+		my $f = $dir.'/'.$fn;
 		next if -f $f;
 		open my $fh, '>', $f or die "open $f: $!";
-		print $fh $INIT_FILES[$i] or die "print $f: $!";
+		print $fh $contents or die "print $f: $!";
 		close $fh or die "close $f: $!";
 	}
 }
@@ -472,10 +483,7 @@ sub done {
 	eval {
 		my $r = delete $self->{in} or die 'BUG: missing {in} when done';
 		print $w "done\n" or wfail;
-		my $pid = delete $self->{pid} or
-				die 'BUG: missing {pid} when done';
-		waitpid($pid, 0) == $pid or die 'fast-import did not finish';
-		$? == 0 or die "fast-import failed: $?";
+		close $r or die "fast-import failed: $?"; # ProcessPipe::CLOSE
 	};
 	my $wait_err = $@;
 	my $nchg = delete $self->{nchg};
@@ -496,8 +504,8 @@ sub atfork_child {
 	}
 }
 
-sub digest2mid ($$) {
-	my ($dig, $hdr) = @_;
+sub digest2mid ($$;$) {
+	my ($dig, $hdr, $fallback_time) = @_;
 	my $b64 = $dig->clone->b64digest;
 	# Make our own URLs nicer:
 	# See "Base 64 Encoding with URL and Filename Safe Alphabet" in RFC4648
@@ -506,7 +514,7 @@ sub digest2mid ($$) {
 	# Add a date prefix to prevent a leading '-' in case that trips
 	# up some tools (e.g. if a Message-ID were a expected as a
 	# command-line arg)
-	my $dt = msg_datestamp($hdr);
+	my $dt = msg_datestamp($hdr, $fallback_time);
 	$dt = POSIX::strftime('%Y%m%d%H%M%S', gmtime($dt));
 	"$dt.$b64" . '@z';
 }
